/**
 *
 * Diplomova prace
 * Stinove techniky na dnesnim hardware a jejich porovnani
 *
 * Michal Toth
 *
 * Fakulta informacnich technologii
 * Vysoke uceni technicke v Brne
 *
 * 2013/2014
 *
 */


#include "chessshader.h"

#include "raycollision.h"

void ChessAnimation::update(int ms)
{
    time+=ms;
    if(time>timeEnd)time=timeEnd;
    float t = (float)(time)/(float)(timeEnd);
    float h=sqrt(qMax(qMin(sinf(t*M_PI),1.0f),0.0f));
    node->translation= (1-t)*posStart+t*posEnd+QVector3D(0,h*height,0);
}


DeadKingAnimation::DeadKingAnimation(){
    timeEnd=4000;
}

void DeadKingAnimation::update(int ms)
{
    time+=ms;
    if(time>timeEnd)time=timeEnd;
    if(time>timeEnd/2){
        float t = (float)(ms)/(float)(timeEnd/2);
        node->rotation+=QVector3D(0,0,t*98);
        node->translation+=QVector3D(0,t*9.7,0);
    }
}


ChessShader::ChessShader(QObject *parent) :
    ShadowCubeMapShader(parent)
{
}



ChessShader::~ChessShader()
{

}

void ChessShader::init()
{
    SceneManager::i()->setActiveScene(Scenes::Chess);
    SceneManager::i()->setActiveBackground(Scenes::Floor);

    ShadowCubeMapShader::init();

    programSquare.addShaderFromSourceFile(QOpenGLShader::Vertex,"shaders/solid.vert");
    programSquare.addShaderFromSourceFile(QOpenGLShader::Fragment,"shaders/solid.frag");
    programSquare.link();

    programPieceHighlight.addShaderFromSourceFile(QOpenGLShader::Vertex,"shaders/Light/phong.vert");
    programPieceHighlight.addShaderFromSourceFile(QOpenGLShader::Fragment,"shaders/Light/phong.frag");
    programPieceHighlight.link();

    Scene* sceneChess = Scenes::getScene(Scenes::Chess);
    QVector<Node*> *nodes = &(sceneChess->root->nodes);

    for(int i=0;i<8;i++){
        pieces.insert(nodes->at(1+i),QPoint(i,1));
        pieces.insert(nodes->at(9+i),QPoint(i,6));
    }
    pieces.insert(nodes->at(17),QPoint(0,0));
    pieces.insert(nodes->at(18),QPoint(7,0));
    pieces.insert(nodes->at(19),QPoint(0,7));
    pieces.insert(nodes->at(20),QPoint(7,7));

    pieces.insert(nodes->at(21),QPoint(1,0));
    pieces.insert(nodes->at(22),QPoint(6,0));
    pieces.insert(nodes->at(23),QPoint(1,7));
    pieces.insert(nodes->at(24),QPoint(6,7));

    pieces.insert(nodes->at(25),QPoint(2,0));
    pieces.insert(nodes->at(26),QPoint(5,0));
    pieces.insert(nodes->at(27),QPoint(2,7));
    pieces.insert(nodes->at(28),QPoint(5,7));

    pieces.insert(nodes->at(29),QPoint(3,0));
    pieces.insert(nodes->at(30),QPoint(3,7));

    pieces.insert(nodes->at(31),QPoint(4,0));
    pieces.insert(nodes->at(32),QPoint(4,7));

    piecesGroup = new Node;
    whitePiecesGroup = new Node;
    blackPiecesGroup = new Node;
    for(int i=1;i<nodes->size();i++){
        piecesGroup->nodes.append(nodes->at(i));
    }

    foreach(Node *n, piecesGroup->nodes) {
        int y = pieces[n].y();
        if(y<=1) whitePiecesGroup->nodes.append(n);
        else blackPiecesGroup->nodes.append(n);
    }

    Node* n = new Node;
    n->geometry = new Geometry(ModelLoader::quad());
    n->scale = QVector3D(18,1,18);
    squaresGroup = new Node();
    for(int y=0;y<8;y++){
        for(int x=0;x<8;x++){
            Node *nn = n->clone();
            nn->translation=QVector3D(70-y*20,3.01,70-x*20);
            boardSquares[y][x]=nn;
            squaresGroup->nodes.append(nn);
            squares.insert(nn,QPoint(x,y));
        }
    }
    squaresGroup->load();


    outGroup = new Node;
    for(int i=0;i<16;i++){
        Node *nn = n->clone();
        nn->translation=QVector3D(-70+(i%8)*20,0,-100-(i/8)*20);
        blackOut[i]=nn;
        outGroup->nodes.append(nn);

        nn=n->clone();
        nn->translation=QVector3D(70-(i%8)*20,0,100+(i/8)*20);
        whiteOut[i]=nn;
        outGroup->nodes.append(nn);
    }

    rayCollision.init();
}

void ChessShader::freeGL(){
    ShadowCubeMapShader::freeGL();
}

void ChessShader::draw(QPainter *painter){

    ShadowCubeMapShader::draw(painter);

    programSquare.bind();
    programSquare.setUniformValue("mvp",mvp);
    programSquare.setUniformValue("m",QMatrix4x4());
    glLineWidth(5);
    glPolygonMode(GL_FRONT_AND_BACK,GL_LINE);

    for(int y=0;y<8;y++){
        for(int x=0;x<8;x++){
            if(selectedPiece!=NULL){
                QPoint p = pieces[selectedPiece];
                QPoint to = QPoint(x,y);
                if(p==to){
                    programSquare.setUniformValue("solidColor",QVector3D(0,1,1));
                    boardSquares[y][x]->draw(&programSquare);
                }else if(chessLogic.canMove(p,to)){
                    if(boardSquares[y][x]==hoverSquare){
                        programSquare.setUniformValue("solidColor",QVector3D(0,1,0));
                    }else{
                        programSquare.setUniformValue("solidColor",QVector3D(1,1,0));
                    }
                    boardSquares[y][x]->draw(&programSquare);
                }
            }
        }
    }
    glPolygonMode(GL_FRONT_AND_BACK,GL_FILL);
    glEnable(GL_DEPTH_TEST);


    if(hoverPiece){
        //glClear(GL_DEPTH_BUFFER_BIT);
        glDepthFunc(GL_LEQUAL);
        glEnable(GL_BLEND);
        glBlendFunc(GL_ONE,GL_ONE);
        programPieceHighlight.bind();
        programPieceHighlight.setUniformValue("mvp",mvp);
        programPieceHighlight.setUniformValue("Ls",QVector3D(0,0,0));
        programPieceHighlight.setUniformValue("Ld",QVector3D(0,0,0));
        programPieceHighlight.setUniformValue("La",QVector3D(0.8,0.8,0.8));

        hoverPiece->draw(&programPieceHighlight);
    }
}

void ChessShader::updateByTime(int ms){

    bool cl = var->clicked;
    var->clicked=false;
    float x = (((float)var->lastMouse.x())/width()-0.5)*2;
    float y = (((float)var->lastMouse.y())/height()-0.5)*2;
    QVector4D a=mvp.inverted()*QVector4D(x,-y,1,1);
    a/=a.w();

    QVector<CollisionResult> * res = rayCollision.collideGeoshader(piecesGroup,Ray(cameraPos,a.toVector3D()));
    if(res->size()>0){
        Node* n = res->at(0).node;
        hoverPiece=n;

        hoverSquare=board(pieces[n]);

        if(cl){
            qDebug() << "whitePiecesGroup"<< whitePiecesGroup->nodes.contains(n);

            if(selectedPiece==n){
                selectedPiece=NULL;
            }else if(whitePiecesGroup->nodes.contains(n)==chessLogic.isWhiteOnMove()){
                selectedPiece=n;
            }
        }
    }else{
        hoverPiece=NULL;

        rayCollision.setCheckBoundingBoxFirst(false);
        QVector<CollisionResult> * res2 = rayCollision.collide(squaresGroup,Ray(cameraPos,a.toVector3D()));
        rayCollision.setCheckBoundingBoxFirst(true);
        if(res2->size()>0){
            hoverSquare = res2->at(0).node;
        }else{
            hoverSquare=NULL;
        }
    }

    if(!animations.empty()){
        for(int i=0;i<animations.size();i++){
            Animation *an = animations.at(i);
            an->update(ms);
            if(an->isFinished()){
                animations.remove(i--);
                delete an;
            }
        }
    }
    if(cl && !chessLogic.isCheckMate()){
        if(selectedPiece!=NULL && hoverSquare!=NULL){
            QPoint from = pieces[selectedPiece];
            QPoint to = squares[hoverSquare];
            if(chessLogic.canMove(from,to)){

                if(pieces.values().contains(to)){
                    Node* targetPiece=pieces.key(to);

                    ChessAnimation *an = new ChessAnimation;
                    an->node=targetPiece;
                    an->height=30;
                    an->timeEnd=2000;
                    an->posStart = targetPiece->translation;

                    if(whitePiecesGroup->nodes.contains(targetPiece)){
                        an->posEnd=whiteOut[whiteOutCounter++]->translation;
                    }else{
                        an->posEnd=blackOut[blackOutCounter++]->translation;
                    }

                    animations.append(an);

                    piecesGroup->nodes.remove(
                                piecesGroup->nodes.indexOf(targetPiece));
                    pieces.remove(targetPiece);
                }

                ChessAnimation *an = new ChessAnimation;
                an->node=selectedPiece;
                an->height=30;
                an->timeEnd=2000;
                an->posStart = selectedPiece->translation;
                an->posEnd=hoverSquare->translation;

                animations.append(an);

                pieces.insert(selectedPiece,to);

                chessLogic.move(from,to);

                hoverSquare=NULL;
                selectedPiece=NULL;
                hoverPiece=NULL;

                if(chessLogic.isCheckMate()){
                    DeadKingAnimation* dka = new DeadKingAnimation;
                    QPoint p=chessLogic.find(chessLogic.isWhiteOnMove()?
                                                  ChessLogic::WHITE_KING:
                                                  ChessLogic::BLACK_KING);
                    dka->node=pieces.key(p);
                    animations.append(dka);
                }
            }
        }
    }


}


